#!   /usr/bin/env   python
#    coding: utf8
##_________________________________________________________________________________________________
##                                                                                                |
##                                           |TDC PTS|                                            |
##                                                                                                |
##                                         CERN,BE/CO-HT                                          |
##________________________________________________________________________________________________|

##-------------------------------------------------------------------------------------------------
##                                                                                                |
##                                   TDC ACAM ERROR FLAG test                                     |
##                                                                                                |
##-------------------------------------------------------------------------------------------------
##                                                                                                |
## Description  Testing of the TDC_ERR signal coming from the ACAM.                               |
##              Through the configuration register 11, the ERR flag is configured to get activated|
##              upon overflow on the Hit FIFOs; an overflow occurs when the timestamps arrival    |
##              arrival rate is higher than 31.25 MHz.                                            |
##              The test is using the Pendulum as pulse generator that in collaboration with the  |
##              Timetech Fanout and the USB-interconnection-box is sending pulses at 50 MHz to two|
##              TDC channels simultaneously and is monitoring the TDC_ERR signal.                 |
##                                                                                                |
##              The TDC_ERR signal can be accessed at PCIe address: 0xC0048                       |
##              The TDC_ERR register can be reset by writing '1' at PCIe address: 0xC004C         |
##                                                                                                |
##                                                                                                |
## FW to load   tdc_allothertests.bit                                                             |
## Authors      Evangelia Gousiou (Evangelia.Gousiou@cern.ch)                                     |
## Website      http://www.ohwr.org/projects/pts                                                  |
## Date         11/01/2013                                                                        |
##-------------------------------------------------------------------------------------------------

##-------------------------------------------------------------------------------------------------
##                               GNU LESSER GENERAL PUBLIC LICENSE                                |
##                              ------------------------------------                              |
## This source file is free software; you can redistribute it and/or modify it under the terms of |
## the GNU Lesser General Public License as published by the Free Software Foundation; either     |
## version 2.1 of the License, or (at your option) any later version.                             |
## This source is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;       |
## without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.      |
## See the GNU Lesser General Public License for more details.                                    |
## You should have received a copy of the GNU Lesser General Public License along with this       |
## source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html                     |
##-------------------------------------------------------------------------------------------------

##-------------------------------------------------------------------------------------------------
##                                            Import                                             --
##-------------------------------------------------------------------------------------------------


# Import system modules
import sys
import time
import os
import math
from datetime import datetime

# Add common modules location tp path
sys.path.append('../../../../pts/')
sys.path.append('../../../../pts/gnurabbit/python/')
sys.path.append('../../../../pts/common/')
sys.path.append('../../../../pts/common/usb_box')
sys.path.append('../../../../pts/common/cp210x')

# Import common modules
from ptsexcept import *
import rr
import csr
from cp210x_eeprom import *
from PendulumCNT91 import *
from calibr_box import *
from find_usb_tty import *
import fmc_tdc

# Pendulum
PENDULUM           = "/dev/usbtmc1"
USB_DEVICE         = "/dev/ttyUSB0"
RS232_BAUD         = 57600

def pendulum_output_on(meas):
    meas.write("OUTP:TYPE PULS")
    meas.write("SOUR:PULS:PER  0.00000002")
    meas.write("SOUR:PULS:WIDT 0.00000001")
    return (1)

def pendulum_output_off(meas):
    meas.write("OUTP:TYPE OFF")
    return (1)

# USB-interconnection-box
BOX_USB_VENDOR_ID = 0x10c4 # Cygnal Integrated Products, Inc.
BOX_USB_DEVICE_ID = 0xea60 # CP210x Composite Device


##-------------------------------------------------------------------------------------------------
##                                             Main                                              --
##-------------------------------------------------------------------------------------------------

def main (default_directory='.'):

    start_test_time = time.time()

#-----> PCIe address, firmware, loader constants
    FMC_TDC_ADDR             = '1a39:0004/1a39:0004@000B:0000'
    FMC_TDC_BITSTREAM_PATH   = '../firmwares/tdc_allothertests.bit'
    FPGA_LOADER_PATH         = '../../../../pts/gnurabbit/user/fpga_loader'

#-----> FPGA constants
    FMC_TDC_CHANNEL_NB  = 5
    ERRFLAG_BASE        = 0xC0000
    ERRFLAG_REG         = 0x48
    RST_ERRFLAG_REG     = 0x4C


#-----> FPGA initializations   
    # SPEC object declaration
    spec = rr.Gennum()

    # Bind SPEC object to FMC TDC card
    print "\n-------------------------------------------------------------------"
    print "----------------------------- FMC TDC -----------------------------"
    print "-------------------- TDC ACAM ERROR FLAG test ---------------------\n"
    print ("Testing of the ACAM TDC_ERR signal. The ERR flag is configured to\n"
           "get activated upon overflow on the Hit FIFOs; an overflow occurs\n"
           "when the timestamps arrival rate is higher than 31.25 MHz.\n"
           "The test is using the Pendulum as pulse generator sending pulses\n"
           "to two TDC channels simultaneously at a  rate of 50 MHz and\n"
           "is monitoring the TDC_ERR signal.\n")

    print "\n_______________________________Info______________________________\n"
    print "FMC TDC address to parse: %s"%(FMC_TDC_ADDR)
    for name, value in spec.parse_addr(FMC_TDC_ADDR).iteritems():
        print "%s:0x%04X"%(name, value)
    spec.bind(FMC_TDC_ADDR)
    ERRFLAG_BASE = csr.CCSR(spec, ERRFLAG_BASE)

    # Load FMC TDC firmware
    print "\n_________________________Initialisations_________________________\n"
    print "Loading FMC TDC firmware...",
    firmware_loader = os.path.join(default_directory, FPGA_LOADER_PATH)
    bitstream = os.path.join(default_directory, FMC_TDC_BITSTREAM_PATH)
    os.system(firmware_loader + ' ' + bitstream)
    time.sleep(1)
    firmware_loader = os.path.join(default_directory, FPGA_LOADER_PATH)
    bitstream = os.path.join(default_directory, FMC_TDC_BITSTREAM_PATH)
    os.system(firmware_loader + ' ' + bitstream)
    time.sleep(1)
    print "Firmware loaded!"

    # TDC object declaration
    tdc = fmc_tdc.CFMCTDC(spec)

    # Check bitsteam type
    bitstream_type = tdc.read_bitstream_type()
    if(bitstream_type == 0xFFFFFFFF):
        msg = ("FATAL ERROR: No access to TDC core")
        os.close(spec.fd)
        print (msg)
        raise PtsCritical (msg)
    if(bitstream_type != 0x1):
        msg = ("FATAL ERROR: Wrong bitstream type loaded")
        #os.close(spec.fd)
        print (msg)
        #raise PtsCritical (msg)
    else:
        print('Access to TDC core OK')


#-----> ACAM configuration
    # Configure ACAM and TDC core registers
    print "\n__________________________Configuration__________________________\n"
    tdc.config_acam()
    time.sleep(1)
    tdc.reset_acam()
    acam_status_test = tdc.read_acam_status()-0xC4000800
    if acam_status_test == 0:
        print "ACAM IC8: Status register OK"
    else:
        msg = ("ERROR: ACAM IC8: No communication")
        os.close(spec.fd)
        print (msg)
        raise PtsError (msg)
    acam_readback_regs = []
    acam_readback_regs = tdc.readback_acam_config()
    for i in range(len(tdc.ACAM_READBACK_REGS)): 
        if (acam_readback_regs[i] == tdc.ACAM_READBACK_REGS[i]):
            print "ACAM IC8: reg 0x%02X: 0x%08X OK"%(tdc.ACAM_READBACK_ADDR[i], acam_readback_regs[i])
        else:
            msg = ("ERROR: ACAM IC8: Configuration registers failure; reg 0x%2X: received 0x%8X, expected 0x%8X"%(tdc.ACAM_READBACK_ADDR[i], acam_readback_regs[i], tdc.ACAM_READBACK_REGS[i]))
            os.close(spec.fd)
            print (msg)
            raise PtsError (msg)


#-----> TDC configuration
    # Enable terminations of channels 1,2,3,4,5
    tdc.enable_channels()
    tdc.channel_term(1, 1)
    tdc.channel_term(2, 1)
    tdc.channel_term(3, 1)
    tdc.channel_term(4, 1)
    tdc.channel_term(5, 1)

    # Enable timestamps acquisition (TDC_START_FPGA pulse sent at this step) 
    tdc.start_acq()

    # Check Effflag register
    ERRFLAG_BASE.wr_reg(RST_ERRFLAG_REG,0x1)
    ERRFLAG_BASE.wr_reg(RST_ERRFLAG_REG,0x0)
    acam_errflag_init = ERRFLAG_BASE.rd_reg(ERRFLAG_REG)
    if acam_errflag_init != 0:
        msg = ("ACAM Error Flag initial value: 0x%08X" %(acam_errflag_init))
        os.close(spec.fd)
        print (msg)
        raise PtsError (msg) 
    else:
        print ("ACAM Error Flag initial value: 0x%08X OK" %(acam_errflag_init))


#-----> USB-interconnection-box
    # Communication and setup of Interconnection box
    print "\n_____________________USB interconnection box______________________\n"
    box_tty = CttyUSB().find_usb_tty(BOX_USB_VENDOR_ID, BOX_USB_DEVICE_ID)
    print("Device: %s" % box_tty[0])
    box = CCalibr_box(box_tty[0])
    box.select_trigger_loopback(0)
    box.select_output(4)


#-----> Pendulum
    # Communication with Pendulum
    print "\n__________________________USB Pendulum_____________________________\n"
    try:
        meas = PendulumCNT91(PENDULUM)
    except OSError:
        raise PtsError('CNT-91 is not switched on or badly connected, aborting.')

    # Pendulum output on
    a = pendulum_output_on(meas)

    # Pendulum sending pulses
    time.sleep(5)

    # Pendulum outsput off
    a = pendulum_output_off(meas)


#-----> Check Errflag
    # Check if timestamps arrived
    print "\n______________________ACAM ErrorFlag test________________________\n"
    print "The time now is                   : %s" %datetime.now()
    print "The temperature now is            : %3.3f°C" %tdc.mezz_get_temp()
    print "TDC overflows counter             : %d"%(tdc.get_overflow_counter())

    if ((ERRFLAG_BASE.rd_reg(ERRFLAG_REG)) != 0):
        print ("ACAM Error Flag                   : 0x%08X OK!" %(ERRFLAG_BASE.rd_reg(ERRFLAG_REG)))
    else:
        msg = ("ERROR: ACAM IC8: Error Flag: not activated as expected:  0x%08X")%(ERRFLAG_BASE.rd_reg(ERRFLAG_REG))
        os.close(spec.fd)
        print (msg)
        raise PtsError (msg) 

    a = pendulum_output_off(meas)


    print "\n\n--------------------------------"
    end_test_time = time.time()
    print "Test elapsed time: %.2f seconds\n\n\n" % (end_test_time-start_test_time)


if __name__ == '__main__' :
    main()
